@@ -5,11 +5,11 @@ Parse command line arguments, invoke embody device.
 import argparse
 import logging
 import sys
-import time
 
-from . import __version__
-from .embodyserial import EmbodySerial
-from .helpers import EmbodySendHelper
+from embodyserial import __version__
+from embodyserial.embodyserial import EmbodySerial
+from embodyserial.helpers import EmbodySendHelper
+from embodyserial.listeners import FileDownloadListener
 
 
 get_attributes_dict: dict[str, str] = {
@@ -63,8 +63,11 @@ def main(args=None):
             __list_files(send_helper)
             exit(0)
         elif parsed_args.download_file:
-            return __download_file(
-                parsed_args.download_file, embody_serial, send_helper
+            __download_file(parsed_args.download_file, embody_serial, send_helper)
+            exit(0)
+        elif parsed_args.download_file_with_delay:
+            __download_file(
+                parsed_args.download_file_with_delay, embody_serial, send_helper, 0.01
             )
             exit(0)
         elif parsed_args.download_files:
@@ -95,7 +98,12 @@ def main(args=None):
 
 def __get_all_attributes(send_helper):
     for attrib in get_attributes_dict.keys():
-        print(f"{attrib}: {getattr(send_helper, get_attributes_dict.get(attrib))()}")
+        sys.stdout.write(f"{attrib}: ")
+        sys.stdout.flush()
+        try:
+            print(getattr(send_helper, get_attributes_dict.get(attrib))())
+        except Exception as e:
+            print(f"Error: {e}")
 
 
 def __list_files(send_helper):
@@ -118,7 +126,10 @@ def __download_files(embody_serial: EmbodySerial, send_helper: EmbodySendHelper)
 
 
 def __download_file(
-    file_name: str, embody_serial: EmbodySerial, send_helper: EmbodySendHelper
+    file_name: str,
+    embody_serial: EmbodySerial,
+    send_helper: EmbodySendHelper,
+    delay: float = 0.0,
 ):
     filtered_files: list[tuple[str, int]] = [
         tup for tup in send_helper.get_files() if tup[0] == file_name
@@ -126,23 +137,57 @@ def __download_file(
     if not filtered_files or len(filtered_files) == 0:
         print(f"Unknown file name {file_name}")
         return
-    __do_download_file(filtered_files[0], embody_serial, send_helper)
+    __do_download_file(filtered_files[0], embody_serial, send_helper, delay)
+
+
+def _show_cli_progress_bar(progress: float, total: int, kbps: float):
+    bar_length = 20
+    percent = progress / 100
+    hashes = "#" * int(round(percent * bar_length))
+    spaces = " " * (bar_length - len(hashes))
+    sys.stdout.write(
+        "\rProgress: [{}] {}% ({} kbps)".format(
+            hashes + spaces, int(round(percent * 100)), int(round(kbps))
+        )
+    )
+    sys.stdout.flush()
 
 
 def __do_download_file(
-    file: tuple[str, int], embody_serial: EmbodySerial, send_helper: EmbodySendHelper
+    file: tuple[str, int],
+    embody_serial: EmbodySerial,
+    send_helper: EmbodySendHelper,
+    delay: float = 0.0,
 ):
     print(f"Downloading: {file[0]}")
-    start = time.time()
-    try:
-        downloaded_file = embody_serial.download_file(file_name=file[0], size=file[1])
-        end = time.time()
-        print(
-            f"{file[0]} downloaded to {downloaded_file} ({round(file[1]/1024,2)}KB)"
-            f" @ ({round((file[1]/1024)/(end-start),2)}KB/s)"
-        )
-    except Exception as e:
-        print(f"Unable to download file: {e}")
+
+    class _DownloadListener(FileDownloadListener):
+        download_invocation_count = 0
+
+        def on_file_download_progress(
+            self, original_file_name: str, size: int, progress: float, kbps: float
+        ) -> None:
+            """Display progress in cli."""
+            if self.download_invocation_count % 10 == 0 or progress == 100:
+                _show_cli_progress_bar(progress, size, kbps)
+            self.download_invocation_count += 1
+
+        def on_file_download_complete(
+            self, original_file_name: str, path: str, kbps: float
+        ) -> None:
+            """Process file download completion."""
+            print(f" {original_file_name} downloaded to {path} (@{round(kbps)} kbps)")
+
+        def on_file_download_failed(
+            self, original_file_name: str, error: Exception
+        ) -> None:
+            """Process file download failure."""
+            print(f" {original_file_name} failed to download: {error}")
+
+    listener = _DownloadListener()
+    embody_serial.download_file(
+        file_name=file[0], size=file[1], download_listener=listener, delay=delay
+    )
 
 
 def __get_args(args):
@@ -165,15 +210,7 @@ def __get_parser():
         choices=log_levels,
         default="WARNING",
     )
-    parser.add_argument(
-        "--channel",
-        help="Use serial or ble",
-        choices=["serial", "ble"],
-        default="serial",
-    )
-    parser.add_argument(
-        "--device", help="Device name (serial or ble name)", default=None
-    )
+    parser.add_argument("--device", help="Serial port name", default=None)
     parser.add_argument(
         "--get", help="Get attribute", choices=get_attributes_dict.keys(), default=None
     )
@@ -186,6 +223,12 @@ def __get_parser():
     parser.add_argument(
         "--download-file", help="Download specified file", type=str, default=None
     )
+    parser.add_argument(
+        "--download-file-with-delay",
+        help="Download specified file with simulated delay",
+        type=str,
+        default=None,
+    )
     parser.add_argument(
         "--download-files", help="Download all files", action="store_true", default=None
     )
