@@ -13,6 +13,7 @@ from abc import ABC
 from abc import abstractmethod
 from concurrent.futures import ThreadPoolExecutor
 from concurrent.futures import TimeoutError
+from dataclasses import dataclass
 from typing import Optional
 
 import serial
@@ -24,11 +25,12 @@ from serial.serialutil import SerialBase
 from serial.serialutil import SerialException
 from serial.tools import list_ports_common
 
-from .exceptions import CrcError
-from .exceptions import MissingResponseError
-from .listeners import ConnectionListener
-from .listeners import MessageListener
-from .listeners import ResponseMessageListener
+from embodyserial.exceptions import CrcError
+from embodyserial.exceptions import MissingResponseError
+from embodyserial.listeners import ConnectionListener
+from embodyserial.listeners import FileDownloadListener
+from embodyserial.listeners import MessageListener
+from embodyserial.listeners import ResponseMessageListener
 
 
 class EmbodySender(ABC):
@@ -109,7 +111,14 @@ class EmbodySerial(ConnectionListener, EmbodySender):
         if not connected:
             self.shutdown()
 
-    def download_file(self, file_name: str, size: int, timeout: int = 300) -> str:
+    def download_file(
+        self,
+        file_name: str,
+        size: int,
+        download_listener: Optional[FileDownloadListener] = None,
+        timeout: int = 300,
+        delay: float = 0.0,
+    ) -> str:
         """Download file from device and write to temporary file.
 
         Raises:
@@ -119,7 +128,9 @@ class EmbodySerial(ConnectionListener, EmbodySender):
         if size == 0:
             return tempfile.NamedTemporaryFile(delete=False).name
         self.send_async(codec.GetFileUart(types.File(file_name)))
-        return self.__reader.download_file(size, timeout)
+        return self.__reader.download_file(
+            file_name, size, download_listener, timeout, delay
+        )
 
     @staticmethod
     def __find_serial_port() -> str:
@@ -224,6 +235,17 @@ class _MessageSender(ResponseMessageListener):
             return None
 
 
+@dataclass
+class _FileDownload:
+    file_size: int = -1
+    file_timeout: Optional[int] = None
+    original_file_name: Optional[str] = None
+    file_name: Optional[str] = None
+    file_error: Optional[Exception] = None
+    file_download_listener: Optional[FileDownloadListener] = None
+    file_delay: float = 0.0
+
+
 class _ReaderThread(threading.Thread):
     """Implement a serial port read loop and dispatch incoming messages to subscribers/listeners.
 
@@ -237,44 +259,68 @@ class _ReaderThread(threading.Thread):
         self.daemon = True
         self.setName("reader")
         self.__serial = serial_instance
+        self.__read_timeout: Optional[int] = None
+        if hasattr(self.__serial, "timeout"):
+            self.__read_timeout = serial_instance.timeout
+
         self.__message_listener_executor = ThreadPoolExecutor(
             max_workers=1, thread_name_prefix="rcv-worker"
         )
         self.__response_message_listener_executor = ThreadPoolExecutor(
             max_workers=1, thread_name_prefix="rsp-worker"
         )
+        self.__file_download_listener_executor = ThreadPoolExecutor(
+            max_workers=1, thread_name_prefix="dwnld-worker"
+        )
         self.__message_listeners: list[MessageListener] = []
         self.__response_message_listeners: list[ResponseMessageListener] = []
         self.__connection_listeners: list[ConnectionListener] = []
         self.__file_mode = False
-        self.__file_size = -1
-        self.__file_timeout = 0
-        self.__file_name: Optional[str] = None
-        self.__file_error: Optional[Exception] = None
+        self.__f: Optional[_FileDownload] = None
         self.__file_event = threading.Event()
         self.alive = True
 
-    def download_file(self, size: int, timeout: int = 300) -> str:
+    def download_file(
+        self,
+        original_file_name: str,
+        size: int,
+        download_listener: Optional[FileDownloadListener] = None,
+        timeout: int = 300,
+        delay=0.0,
+    ) -> str:
         """Set reader in file mode and read file."""
+        if hasattr(self.__serial, "timeout"):
+            self.__serial.timeout = 10
         self.__reset_file_mode()
-        self.__file_timeout = timeout
-        self.__file_size = size
+        f = _FileDownload(
+            file_size=size,
+            file_timeout=timeout,
+            original_file_name=original_file_name,
+            file_download_listener=download_listener,
+            file_delay=delay,
+        )
+        self.__f = f
         self.__file_mode = True
         try:
-            if not self.__file_event.wait(timeout) or not self.__file_name:
+            if not self.__file_event.wait(timeout):
                 raise MissingResponseError("No file received within timeout")
-            if self.__file_error:
-                raise self.__file_error
-            return self.__file_name
+            if self.__f.file_error:
+                raise self.__f.file_error
+            if not self.__f.file_name:
+                raise MissingResponseError("No file received")
+            return self.__f.file_name
+        except Exception as e:
+            self.__async_notify_file_download_failed(f, e)
+            raise e
         finally:
             self.__reset_file_mode()
+            if hasattr(self.__serial, "timeout") and self.__read_timeout:
+                self.__serial.timeout = self.__read_timeout
 
     def __reset_file_mode(self) -> None:
         self.__file_event.clear()
-        self.__file_size = -1
-        self.__file_name = None
-        self.__file_error = None
         self.__file_mode = False
+        self.__f = None
 
     def stop(self) -> None:
         """Stop the reader thread"""
@@ -287,6 +333,9 @@ class _ReaderThread(threading.Thread):
         self.__response_message_listener_executor.shutdown(
             wait=False, cancel_futures=False
         )
+        self.__file_download_listener_executor.shutdown(
+            wait=False, cancel_futures=False
+        )
         self.join(2)
 
     def run(self) -> None:
@@ -299,9 +348,8 @@ class _ReaderThread(threading.Thread):
                 if not raw_header or len(raw_header) < 3:
                     logging.info("Interrupted. Exiting reader thread.")
                     break
-                if self.__file_mode:
-                    self.__read_file(raw_header)
-                    self.__reset_file_mode()
+                if self.__file_mode and self.__f:
+                    self.__read_file(raw_header, self.__f)
                 else:
                     self.__read_protocol_message(raw_header)
             except serial.SerialException:
@@ -314,45 +362,99 @@ class _ReaderThread(threading.Thread):
         self.alive = False
         self.__notify_connection_listeners(connected=False)
 
-    def __read_file(self, first_bytes: bytes) -> None:
-        buffer_size = 1024
-        remaining_size = self.__file_size - len(first_bytes)
-        calculated_crc = crc.crc16(data=first_bytes)
-        tmp = tempfile.NamedTemporaryFile(delete=False, buffering=buffer_size)
+    def __read_file(self, first_bytes: bytes, f: _FileDownload) -> None:
+        buffer_size = 2048
+        remaining_size = f.file_size - len(first_bytes)
         start = time.time()
-        tmp.write(first_bytes)
+        in_memory_buffer = bytearray(first_bytes)
+        loop_count = 0
         try:
             while remaining_size > 0 and self.__serial.is_open:
                 chunk = self.__serial.read(min(buffer_size, remaining_size))
                 if not chunk:
                     raise MissingResponseError("File download failed")
-                calculated_crc = crc.crc16(data=chunk, existing_crc=calculated_crc)
-                tmp.write(chunk)
+                in_memory_buffer.extend(chunk)
                 remaining_size -= len(chunk)
-                logging.debug(
-                    f"Read file. Remaining: {remaining_size} of {self.__file_size}"
-                    f" bytes (in waiting: {self.__serial.in_waiting} bytes)"
-                )
-                time.sleep(0.001)
-                if time.time() - start > self.__file_timeout:
-                    raise TimeoutError(
-                        f"Reading file took too long. Read {self.__file_size - remaining_size} bytes"
+                now = time.time()
+                if loop_count % 20 == 0:
+                    self.__async_notify_file_download_in_progress(
+                        f,
+                        f.file_size,
+                        round(((f.file_size - remaining_size) / f.file_size) * 100),
+                        round(((f.file_size - remaining_size) / 1024) / (now - start)),
                     )
+                loop_count += 1
+                if f.file_timeout and now - start > f.file_timeout:
+                    raise TimeoutError(
+                        f"Reading file took too long. Read {f.file_size - remaining_size} bytes"
+                    )
+                if f.file_delay:
+                    time.sleep(f.file_delay)
             raw_crc_received = self.__serial.read(2)
             end = time.time()
-            logging.info(
-                f"Read {round(self.__file_size/1024,2)}KB in {end-start} secs - {round((self.__file_size/1024)/(end-start),2)}KB/s"
+            self.__async_notify_file_download_in_progress(
+                f,
+                f.file_size,
+                100,
+                round((f.file_size / 1024) / (end - start), 2),
+            )
+            logging.debug(
+                f"Read {round(f.file_size/1024,2)}KB in {end-start} secs "
+                f"- {round((f.file_size/1024)/(end-start),2)}KB/s"
             )
             (crc_received,) = struct.unpack(">H", raw_crc_received)
+            calculated_crc = crc.crc16(data=in_memory_buffer)
             if not crc_received == calculated_crc:
-                self.__file_error = CrcError(
-                    f"Invalid crc - expected {hex(crc_received)}, received {hex(calculated_crc)}"
+                f.file_error = CrcError(
+                    f"Invalid crc - expected {hex(crc_received)}, received/calculated {hex(calculated_crc)}"
                 )
-            else:
-                self.__file_name = tmp.name
-            self.__file_event.set()
-        finally:
+                self.__async_notify_file_download_failed(f, f.file_error)
+                return
+            tmp = tempfile.NamedTemporaryFile(delete=False)
+            tmp.write(in_memory_buffer)
+            tmp.flush()
             tmp.close()
+            f.file_name = tmp.name
+            self.__async_notify_file_download_completed(
+                f, round((f.file_size / 1024) / (end - start), 2)
+            )
+        except Exception as e:
+            f.file_error = e
+            self.__async_notify_file_download_failed(f, e)
+        finally:
+            self.__file_event.set()
+
+    def __async_notify_file_download_in_progress(
+        self, f: _FileDownload, size: int, progress: float, kbps: float
+    ):
+        if f.file_download_listener and f.original_file_name:
+            self.__file_download_listener_executor.submit(
+                _ReaderThread.__notify_file_download_progress,
+                f.file_download_listener,
+                f.original_file_name,
+                size,
+                progress,
+                kbps,
+            )
+
+    def __async_notify_file_download_completed(self, f: _FileDownload, kbps: float):
+        if f.file_download_listener and f.original_file_name and f.file_name:
+            self.__file_download_listener_executor.submit(
+                _ReaderThread.__notify_file_download_complete,
+                f.file_download_listener,
+                f.original_file_name,
+                f.file_name,
+                kbps,
+            )
+
+    def __async_notify_file_download_failed(self, f: _FileDownload, error: Exception):
+        if f.file_download_listener and f.original_file_name:
+            self.__file_download_listener_executor.submit(
+                _ReaderThread.__notify_file_download_failed,
+                f.file_download_listener,
+                f.original_file_name,
+                error,
+            )
 
     def __read_protocol_message(self, raw_header: bytes) -> None:
         """Read next message from input."""
@@ -447,6 +549,43 @@ class _ReaderThread(threading.Thread):
                 f"Error notifying connection listener: {str(e)}", exc_info=True
             )
 
+    @staticmethod
+    def __notify_file_download_progress(
+        listener: FileDownloadListener,
+        original_file_name: str,
+        size: int,
+        progress: float,
+        kbps: float,
+    ) -> None:
+        try:
+            listener.on_file_download_progress(original_file_name, size, progress, kbps)
+        except Exception as e:
+            logging.warning(
+                f"Error notifying file download listener: {str(e)}", exc_info=True
+            )
+
+    @staticmethod
+    def __notify_file_download_complete(
+        listener: FileDownloadListener, original_file_name: str, path: str, kbps: float
+    ) -> None:
+        try:
+            listener.on_file_download_complete(original_file_name, path, kbps)
+        except Exception as e:
+            logging.warning(
+                f"Error notifying file download listener: {str(e)}", exc_info=True
+            )
+
+    @staticmethod
+    def __notify_file_download_failed(
+        listener: FileDownloadListener, original_file_name: str, error: Exception
+    ) -> None:
+        try:
+            listener.on_file_download_failed(original_file_name, error)
+        except Exception as e:
+            logging.warning(
+                f"Error notifying file download listener: {str(e)}", exc_info=True
+            )
+
 
 if __name__ == "__main__":
     """Main method for demo and testing"""
